from pyrosm.data_manager import get_osm_data
from pyrosm.frames import prepare_geodataframe
from pyrosm.utils import validate_custom_filter
import warnings


def get_building_data(
    node_coordinates,
    way_records,
    relations,
    tags_as_columns,
    custom_filter,
    bounding_box,
):
    # If custom_filter has not been defined, initialize with default
    if custom_filter is None:
        custom_filter = {"building": [True]}
    else:
        # Check that the custom filter is in correct format
        custom_filter = validate_custom_filter(custom_filter)

        # Ensure that the "building" tag exists
        if "building" not in custom_filter.keys():
            custom_filter["building"] = [True]

    # Call signature for fetching buildings
    nodes, ways, relation_ways, relations = get_osm_data(
        node_arrays=None,
        way_records=way_records,
        relations=relations,
        tags_as_columns=tags_as_columns,
        data_filter=custom_filter,
        filter_type="keep",
        osm_keys=None,
    )

    # If there weren't any data, return empty GeoDataFrame
    if nodes is None and ways is None and relations is None:
        warnings.warn(
            "Could not find any building elements for given area.",
            UserWarning,
            stacklevel=2,
        )
        return None

    # Prepare GeoDataFrame
    gdf = prepare_geodataframe(
        nodes,
        node_coordinates,
        ways,
        relations,
        relation_ways,
        tags_as_columns,
        bounding_box,
    )
    return gdf



from pyrosm.config.default_tags import *
from pyrosm.config.osm_filters import get_osm_filter


class NetworkFilter:
    driving = get_osm_filter("driving")
    driving_psv = get_osm_filter("driving+psv")
    walking = get_osm_filter("walking")
    cycling = get_osm_filter("cycling")


class Tags:
    # Tags object contains configuration about the default
    # tag combinations (key:value) that will be kept as columns
    # in the resulting GeoDataFrame. All other possible tags are
    # inserted into a JSON that is stored in "tags" column.
    # These follow more or less OSM Wiki documentation:
    # https://wiki.openstreetmap.org/wiki/Map_Features
    available = [
        "aerialway",
        "aeroway",
        "amenity",
        "boundary",
        "building",
        "craft",
        "emergency",
        "geological",
        "highway",
        "historic",
        "landuse",
        "leisure",
        "natural",
        "office",
        "power",
        "public_transport",
        "railway",
        "route",
        "place",
        "shop",
        "tourism",
        "waterway",
    ]

    aerialway = aerialway_columns
    aeroway = aeroway_columns
    amenity = amenity_columns
    boundary = boundary_columns
    building = building_columns
    craft = craft_columns
    emergency = emergency_columns
    geological = geological_columns
    highway = highway_columns
    historic = historic_columns
    landuse = landuse_columns
    leisure = leisure_columns
    natural = natural_columns
    office = office_columns
    power = power_columns
    public_transport = public_transport_columns
    railway = railway_columns
    route = route_columns
    place = place_columns
    shop = shop_columns
    tourism = tourism_columns
    waterway = waterway_columns

    # Basic info tags
    _basic_tags = basic_info_tags


class Conf:
    network_filters = NetworkFilter()
    tags = Tags()
    _possible_network_filters = [a for a in network_filters.__dir__() if "__" not in a]
    _possible_network_filters += ["all", "driving+service"]

    # One way tags
    oneway_values = ["yes", "true", "1", "-1", "T", "F"]



cpdef create_nodes_gdf(node_dict_list)
cpdef create_gdf(data_records, geometry_array)
cpdef prepare_way_gdf(node_coordinates, ways, parse_network, calculate_seg_lengths)
cpdef prepare_node_gdf(nodes)
cpdef prepare_geodataframe(nodes,
                           node_coordinates,
                           ways,
                           relations,
                           relation_ways,
                           tags_as_columns,
                           bounding_box,
                           parse_network=*,
                           calculate_seg_lengths=*)


cpdef _create_igraph(nodes, edges, from_id_col, to_id_col, node_id_col)
cpdef _create_nxgraph(nodes, edges, from_id_col, to_id_col, node_id_col)
cpdef _create_pdgraph(nodes, edges, from_id_col, to_id_col, weight_cols)
cpdef generate_directed_edges(edges, direction, from_id_col, to_id_col, force_bidirectional)




import pandas as pd
import geopandas as gpd
import numpy as np
from pyrosm._arrays cimport concatenate_dicts_of_arrays
from pyrosm.geometry cimport _create_point_geometries
from pyrosm.geometry cimport create_way_geometries
from pyrosm.relations import prepare_relations
from shapely.geometry import box
from pyrosm.data_filter import get_mask_by_osmid, _filter_array_dict_by_indices_or_mask
from pygeos import multilinestrings
from pyrosm.distance import calculate_geom_length, calculate_geom_array_length

cpdef create_nodes_gdf(nodes, osmids_to_keep=None):
    cdef str k
    if isinstance(nodes, list):
        nodes = concatenate_dicts_of_arrays(nodes)

    # Check if nodes should be filtered
    if osmids_to_keep is not None:
        if isinstance(osmids_to_keep, np.ndarray):
            # Get mask for nodeid array
            mask = get_mask_by_osmid(nodes["id"], osmids_to_keep)
            nodes = _filter_array_dict_by_indices_or_mask(nodes, mask)
        else:
            raise ValueError("'indices_to_keep' should be a numpy array.")

    df = pd.DataFrame()
    for k, v in nodes.items():
        df[k] = v
    df['geometry'] = _create_point_geometries(nodes['lon'], nodes['lat'])
    return gpd.GeoDataFrame(df, crs='epsg:4326')

cpdef create_gdf(data_arrays, geometry_array):
    df = create_df(data_arrays)
    df['geometry'] = geometry_array
    return gpd.GeoDataFrame(df, crs='epsg:4326')

cpdef create_df(data_arrays):
    cdef str key
    df = pd.DataFrame()
    for key, data in data_arrays.items():
        # When inserting nodes,
        # those should be converted
        # to lists to avoid block error
        if key == "nodes":
            df[key] = data.tolist()
        else:
            df[key] = data

    return df

cpdef prepare_way_gdf(node_coordinates, ways, parse_network, calculate_seg_lengths):
    if ways is not None:
        geometries, from_ids, to_ids, node_attributes = create_way_geometries(node_coordinates,
                                                                              ways,
                                                                              parse_network
                                                                              )
        # Convert to DataFrame
        way_gdf = create_df(ways)
        way_gdf['osm_type'] = "way"
        way_gdf["geometry"] = geometries

        # In case network is parsed, include way-level length info
        if parse_network and not calculate_seg_lengths:
            # Drop rows without geometry
            way_gdf = way_gdf.dropna(subset=['geometry']).reset_index(drop=True)

            # Create MultiLineStrings and calculate the length
            way_gdf["geometry"] = [multilinestrings(geom) for geom in way_gdf["geometry"]]
            way_gdf["length"] = [calculate_geom_length(geom) for geom in way_gdf["geometry"]]
            way_gdf = gpd.GeoDataFrame(way_gdf, geometry="geometry", crs="epsg:4326")

            # If only edges are requested, clean node_attributes
            node_attributes = None

        # In case network is parsed and requested for graph export,
        # include segment-level length info
        elif parse_network and calculate_seg_lengths:
            # Insert way-level from/to ids
            way_gdf["u"] = from_ids
            way_gdf["v"] = to_ids

            # Drop rows without geometry
            way_gdf = way_gdf.dropna(subset=['geometry']).reset_index(drop=True)

            # Parse segment level from/to-ids
            u = np.concatenate(way_gdf["u"].to_list())
            v = np.concatenate(way_gdf["v"].to_list())

            # Explode multi-geometries
            way_gdf = way_gdf.explode("geometry").reset_index(drop=True)
            way_gdf = gpd.GeoDataFrame(way_gdf, geometry="geometry", crs="epsg:4326")

            # Update from/to-ids
            way_gdf["u"] = u
            way_gdf["v"] = v

            # Calculate the length of the geometries
            way_gdf["length"] = calculate_geom_array_length(way_gdf.geometry.values.data)

        # For cases not related to networks
        else:
            way_gdf = gpd.GeoDataFrame(way_gdf, geometry="geometry", crs="epsg:4326")
            node_attributes = None

    else:
        way_gdf = gpd.GeoDataFrame()
        node_attributes = None

    return way_gdf, node_attributes

cpdef prepare_node_gdf(nodes):
    if nodes is not None:
        # Create GeoDataFrame from nodes
        node_gdf = create_nodes_gdf(nodes)
        node_gdf['osm_type'] = "node"
    else:
        node_gdf = gpd.GeoDataFrame()
    return node_gdf

cpdef prepare_relation_gdf(node_coordinates, relations, relation_ways, tags_as_columns):
    if relations is not None:
        relations = prepare_relations(relations, relation_ways,
                                      node_coordinates,
                                      tags_as_columns)

        relation_gdf = gpd.GeoDataFrame(relations)
        relation_gdf['osm_type'] = "relation"

    else:
        relation_gdf = gpd.GeoDataFrame()
    return relation_gdf

cpdef prepare_geodataframe(nodes, node_coordinates, ways,
                           relations, relation_ways,
                           tags_as_columns, bounding_box,
                           parse_network=False,
                           calculate_seg_lengths=False):

    # Prepare ways
    way_gdf, node_attr = prepare_way_gdf(node_coordinates,
                                         ways,
                                         parse_network,
                                         calculate_seg_lengths)

    # Prepare relation data
    relation_gdf = prepare_relation_gdf(node_coordinates, relations, relation_ways, tags_as_columns)

    # When not parsing the network,
    # nodes should be kept as part of the main output
    if not parse_network:
        # Prepare nodes
        node_gdf = prepare_node_gdf(nodes)
    else:
        node_gdf = pd.DataFrame()

    # Merge all
    gdf = pd.concat([node_gdf, way_gdf, relation_gdf])

    if len(gdf) == 0:
        if parse_network:
            return None, None
        return None

    gdf = gdf.dropna(subset=['geometry']).reset_index(drop=True)

    # When parsing the network with nodes, prepare the nodes frame
    if node_attr is not None:
        node_attr = pd.DataFrame(node_attr)
        node_attr = gpd.GeoDataFrame(node_attr,
                                     crs="epsg:4326",
                                     geometry=gpd.points_from_xy(node_attr["lon"],
                                                                 node_attr["lat"])
                                     ).drop_duplicates("id").reset_index(drop=True)

    # Filter by bounding box if it was used
    if bounding_box is not None:
        if isinstance(bounding_box, list):
            bounding_box = box(*bounding_box)
        # Filter data spatially
        orig_cols = list(gdf.columns)
        filter_gdf = gpd.GeoDataFrame({"geometry": [bounding_box]},
                                      crs="epsg:4326",
                                      index=[0])
        gdf = gpd.sjoin(gdf, filter_gdf, how="inner")
        gdf = gdf[orig_cols].reset_index(drop=True)

        if node_attr is not None:
            orig_node_cols = list(node_attr.columns)
            node_attr = gpd.sjoin(node_attr, filter_gdf, how="inner")
            node_attr = node_attr[orig_node_cols]

    if len(gdf) == 0:
        if parse_network:
            return None, None
        return None

    if parse_network:
        return gdf, node_attr
    return gdf



from pyrosm.data_manager import get_osm_data
from pyrosm.frames import prepare_geodataframe
import warnings


def get_user_defined_data(
    nodes,
    node_coordinates,
    way_records,
    relations,
    tags_as_columns,
    custom_filter,
    osm_keys,
    filter_type,
    keep_nodes,
    keep_ways,
    keep_relations,
    bounding_box,
):

    if not keep_nodes:
        nodes = None

    # If wanting to parse relations but not ways,
    # it is still necessary to parse ways as well at this point
    if keep_ways is False and keep_relations is True:
        pass
    # If ways are not wanted, neither should relations be parsed
    elif not keep_ways:
        way_records = None
        relations = None

    if not keep_relations:
        relations = None

    # Call signature for fetching POIs
    nodes, ways, relation_ways, relations = get_osm_data(
        node_arrays=nodes,
        way_records=way_records,
        relations=relations,
        tags_as_columns=tags_as_columns,
        data_filter=custom_filter,
        filter_type=filter_type,
        osm_keys=osm_keys,
    )

    # If there weren't any data, return empty GeoDataFrame
    if nodes is None and ways is None and relations is None:
        warnings.warn(
            "Could not find any OSM data for given area.", UserWarning, stacklevel=2
        )
        return None

    # Ensure that ways are None if returning those are not requested
    if not keep_ways:
        ways = None

    # Prepare GeoDataFrame
    gdf = prepare_geodataframe(
        nodes,
        node_coordinates,
        ways,
        relations,
        relation_ways,
        tags_as_columns,
        bounding_box,
    )

    return gdf



from pyrosm.data_manager import get_osm_data
from pyrosm.frames import prepare_geodataframe
from pyrosm.utils import validate_custom_filter
import warnings


def get_poi_data(
    nodes,
    node_coordinates,
    way_records,
    relations,
    tags_as_columns,
    custom_filter,
    bounding_box,
):
    # Validate filter
    custom_filter = validate_custom_filter(custom_filter)

    # Call signature for fetching POIs
    nodes, ways, relation_ways, relations = get_osm_data(
        node_arrays=nodes,
        way_records=way_records,
        relations=relations,
        tags_as_columns=tags_as_columns,
        data_filter=custom_filter,
        filter_type="keep",
        osm_keys=None,
    )

    # If there weren't any data, return empty GeoDataFrame
    if nodes is None and ways is None and relations is None:
        warnings.warn(
            "Could not find any POIs for given area.", UserWarning, stacklevel=2
        )
        return None

    # Prepare GeoDataFrame
    gdf = prepare_geodataframe(
        nodes,
        node_coordinates,
        ways,
        relations,
        relation_ways,
        tags_as_columns,
        bounding_box,
    )

    return gdf



from shapely.geos import geos_version_string as shapely_geos_version
from pygeos import geos_capi_version_string
import warnings

# shapely has something like: "3.6.2-CAPI-1.10.2 4d2925d6"
# pygeos has something like: "3.6.2-CAPI-1.10.2"
if not shapely_geos_version.startswith(geos_capi_version_string):
    PYGEOS_SHAPELY_COMPAT = False
else:
    PYGEOS_SHAPELY_COMPAT = True

# python-igraph is an optional dependency
try:
    import igraph

    HAS_IGRAPH = True
except ImportError:
    HAS_IGRAPH = False

# networkx is an optional dependency
try:
    import networkx

    HAS_NETWORKX = True
except ImportError:
    HAS_NETWORKX = False

# pandana is an optional dependency
try:
    import pandana

    HAS_PANDANA = True
except ImportError:
    HAS_PANDANA = False



# BASE
URL = "https://download.bbbike.org/osm/bbbike"
suffix = ".osm.pbf"


class Cities:
    available = [
        "Aachen",
        "Aarhus",
        "Adelaide",
        "Albuquerque",
        "Alexandria",
        "Amsterdam",
        "Antwerpen",
        "Arnhem",
        "Auckland",
        "Augsburg",
        "Austin",
        "Baghdad",
        "Baku",
        "Balaton",
        "Bamberg",
        "Bangkok",
        "Barcelona",
        "Basel",
        "Beijing",
        "Beirut",
        "Berkeley",
        "Berlin",
        "Bern",
        "Bielefeld",
        "Birmingham",
        "Bochum",
        "Bogota",
        "Bombay",
        "Bonn",
        "Bordeaux",
        "Boulder",
        "BrandenburgHavel",
        "Braunschweig",
        "Bremen",
        "Bremerhaven",
        "Brisbane",
        "Bristol",
        "Brno",
        "Bruegge",
        "Bruessel",
        "Budapest",
        "BuenosAires",
        "Cairo",
        "Calgary",
        "Cambridge",
        "CambridgeMa",
        "Canberra",
        "CapeTown",
        "Chemnitz",
        "Chicago",
        "ClermontFerrand",
        "Colmar",
        "Copenhagen",
        "Cork",
        "Corsica",
        "Corvallis",
        "Cottbus",
        "Cracow",
        "CraterLake",
        "Curitiba",
        "Cusco",
        "Dallas",
        "Darmstadt",
        "Davis",
        "DenHaag",
        "Denver",
        "Dessau",
        "Dortmund",
        "Dresden",
        "Dublin",
        "Duesseldorf",
        "Duisburg",
        "Edinburgh",
        "Eindhoven",
        "Emden",
        "Erfurt",
        "Erlangen",
        "Eugene",
        "Flensburg",
        "FortCollins",
        "Frankfurt",
        "FrankfurtOder",
        "Freiburg",
        "Gdansk",
        "Genf",
        "Gent",
        "Gera",
        "Glasgow",
        "Gliwice",
        "Goerlitz",
        "Goeteborg",
        "Goettingen",
        "Graz",
        "Groningen",
        "Halifax",
        "Halle",
        "Hamburg",
        "Hamm",
        "Hannover",
        "Heilbronn",
        "Helsinki",
        "Hertogenbosch",
        "Huntsville",
        "Innsbruck",
        "Istanbul",
        "Jena",
        "Jerusalem",
        "Johannesburg",
        "Kaiserslautern",
        "Karlsruhe",
        "Kassel",
        "Katowice",
        "Kaunas",
        "Kiel",
        "Kiew",
        "Koblenz",
        "Koeln",
        "Konstanz",
        "LaPaz",
        "LaPlata",
        "LakeGarda",
        "Lausanne",
        "Leeds",
        "Leipzig",
        "Lima",
        "Linz",
        "Lisbon",
        "Liverpool",
        "Ljubljana",
        "Lodz",
        "London",
        "Luebeck",
        "Luxemburg",
        "Lyon",
        "Maastricht",
        "Madison",
        "Madrid",
        "Magdeburg",
        "Mainz",
        "Malmoe",
        "Manchester",
        "Mannheim",
        "Marseille",
        "Melbourne",
        "Memphis",
        "MexicoCity",
        "Miami",
        "Moenchengladbach",
        "Montevideo",
        "Montpellier",
        "Montreal",
        "Moscow",
        "Muenchen",
        "Muenster",
        "NewDelhi",
        "NewOrleans",
        "NewYorkCity",
        "Nuernberg",
        "Oldenburg",
        "Oranienburg",
        "Orlando",
        "Oslo",
        "Osnabrueck",
        "Ostrava",
        "Ottawa",
        "Paderborn",
        "Palma",
        "PaloAlto",
        "Paris",
        "Perth",
        "Philadelphia",
        "PhnomPenh",
        "Portland",
        "PortlandME",
        "Porto",
        "PortoAlegre",
        "Potsdam",
        "Poznan",
        "Prag",
        "Providence",
        "Regensburg",
        "Riga",
        "RiodeJaneiro",
        "Rostock",
        "Rotterdam",
        "Ruegen",
        "Saarbruecken",
        "Sacramento",
        "Saigon",
        "Salzburg",
        "SanFrancisco",
        "SanJose",
        "SanktPetersburg",
        "SantaBarbara",
        "SantaCruz",
        "Santiago",
        "Sarajewo",
        "Schwerin",
        "Seattle",
        "Seoul",
        "Sheffield",
        "Singapore",
        "Sofia",
        "Stockholm",
        "Stockton",
        "Strassburg",
        "Stuttgart",
        "Sucre",
        "Sydney",
        "Szczecin",
        "Tallinn",
        "Tehran",
        "Tilburg",
        "Tokyo",
        "Toronto",
        "Toulouse",
        "Trondheim",
        "Tucson",
        "Turin",
        "UlanBator",
        "Ulm",
        "Usedom",
        "Utrecht",
        "Vancouver",
        "Victoria",
        "WarenMueritz",
        "Warsaw",
        "WashingtonDC",
        "Waterloo",
        "Wien",
        "Wroclaw",
        "Wuerzburg",
        "Wuppertal",
        "Zagreb",
        "Zuerich",
    ]

    available.sort()

    # Create data sources
    _sources = {
        city.lower(): {"name": city + suffix, "url": f"{URL}/{city}/{city}{suffix}"}
        for city in available
        if city not in ["NewYorkCity"]
    }
    # Add New York City separately as there is also a state with the same name
    # NewYork ==> NewYorkCity
    _ny = "newyorkcity"
    _ny_src = "NewYork"
    _sources[_ny] = {"name": _ny + suffix, "url": f"{URL}/{_ny_src}/{_ny_src}{suffix}"}

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name.lower()]



# ===============================
# OSM TAGS
# ===============================
# These configurations are used to specify which OSM tags are kept
# by default as columns in DataFrame. Follows more or less the
# OSM documentation and the tags mentioned on the Wiki:
# https://wiki.openstreetmap.org/wiki/Map_Features

# If a tag is NOT mentioned in these lists (below),
# such tag will be kept in a JSON that is stored in a "tag" column.
# The purpose of specifying default tag structure is to speed up the parsing, and
# reduce the complexity in the resulting DataFrame as OSM data can contain numerous
# attributes that are not most relevant for basic user (e.g. "customized" tags).

# ========================
# BASIC INFO TAGS
# ========================
# A list of tags that are relevant for most OSM keys
basic_info_tags = [
    "addr:city",
    "addr:country",
    "addr:full",
    "addr:housenumber",
    "addr:housename",
    "addr:postcode",
    "addr:place",
    "addr:street",
    "email",
    "name",
    "opening_hours",
    "operator",
    "phone",
    "ref",
    "url",
    "website",
    "yes",
]

# ========================
# HIGHWAY TAGS
# ========================

# Default tags to keep as columns with highways
# Mostly based on: https://wiki.openstreetmap.org/wiki/Key:highway
highway_columns = [
    "access",
    "area",
    "bicycle",
    "bicycle_road",
    "bridge",
    "busway",
    "cycleway",
    "est_width",
    "foot",
    "footway",
    "highway",
    "int_ref",
    "junction",
    "lanes",
    "lit",
    "maxspeed",
    "motorcar",
    "motorroad",
    "motor_vehicle",
    "name",
    "oneway",
    "overtaking",
    "path",
    "passing_places",
    "psv",
    "ref",
    "service",
    "segregated",
    "sidewalk",
    "smoothness",
    "surface",
    "tracktype",
    "tunnel",
    "turn",
    "width",
    "winter_road",
    # Other highway tags which are not kept by default
    # (more tags slows down the parsing)
    # =====================================
    # "abutters",
    # "driving_side",
    # "embedded_rails",
    # "ford",
    # "ice_road",
    # "incline",
    # "mtb:scale",
    # "mtb:scale:uphill",
    # "mtb:scale:imba",
    # "mtb:description",
    # "parking:condition",
    # "parking:lane",
    # "sac_scale",
    # "tactile_paving",
    # "traffic_calming",
    # "trail_visibility",
]

# ========================
# BUILDING TAGS
# ========================
# See:
# https://wiki.openstreetmap.org/wiki/Key:building
# https://wiki.openstreetmap.org/wiki/Key:addr

# Default tags to keep as columns with buildings

building_columns = basic_info_tags + [
    "building",
    "amenity",
    "building:flats",
    "building:levels",
    "building:material",
    "building:max_level",
    "building:min_level",
    "building:fireproof",
    "building:use",
    "craft",
    "height",
    "internet_access",
    "landuse",
    "levels",
    "office",
    "operator",
    "shop",
    "source",
    "start_date",
    "wikipedia",
    # Other tags that are not kept by default
    # "addr:conscriptionnumber",
    # "addr:district",
    # "addr:province",
    # "soft_storey",
    # "takeaway",
    # "addr:state",
    # "mml:class"
]

# ========================
# AMENITY TAGS
# ========================
# See:
# https://wiki.openstreetmap.org/wiki/Key:amenity

# Default tags to keep as columns with amenities
amenity_columns = basic_info_tags + [
    "amenity",
    "arts_centre",
    "atm",
    "bank",
    "bicycle_parking",
    "bicycle_rental",
    "bicycle_repair_station",
    "bar",
    "bbq",
    "biergarten",
    "brothel",
    "building",
    "building:levels",
    "bureau_de_change",
    "bus_station",
    "bus_stop",
    "cafe",
    "car_rental",
    "car_repair",
    "car_sharing",
    "car_wash",
    "casino",
    "charging_station",
    "childcare",
    "cinema",
    "clinic",
    "college",
    "dentist",
    "doctors",
    "driving_school",
    "drinking_water",
    "fast_food",
    "ferry_terminal",
    "fire_station",
    "food_court",
    "fountain",
    "fuel",
    "gambling",
    "hospital",
    "ice_cream",
    "internet_access",
    "kindergarten",
    "landuse",
    "language_school",
    "library",
    "music_school",
    "nightclub",
    "nursing_home",
    "office",
    "operator",
    "parking",
    "pharmasy",
    "planetarium",
    "police",
    "post_office",
    "pub",
    "public_bath",
    "rescue_station",
    "restaurant",
    "retirement_home",
    "school",
    "social_centre",
    "social_facility",
    "source",
    "spa",
    "start_date",
    "stripclub",
    "taxi",
    "theatre",
    "university",
    "wikipedia",
    # Other tags
    # ----------
    # "baby_hatch",
    # "levels",
    # "ev_charging",
    # "motorcycle_parking",
    # "parking_entrance",
    # "parking_space",
    # "dive_centre",
    # "dojo",
]

# ========================
# SHOP TAGS
# ========================
# See:
# https://wiki.openstreetmap.org/wiki/Key:shop

# Default tags to keep as columns with amenities
shop_columns = basic_info_tags + [
    "agrarian",
    "alcohol",
    "anime",
    "antiques",
    "appliance",
    "art",
    "atv",
    "baby_goods",
    "bag",
    "bakery",
    "bathroom_furnishing",
    "bed",
    "beverages",
    "bicycle",
    "boat",
    "bookmaker",
    "books",
    "boutique",
    "brewing_supplies",
    "butcher",
    "camera",
    "candles",
    "cannabis",
    "car",
    "car_parts",
    "car_repair",
    "caravan",
    "carpet",
    "charity",
    "cheese",
    "chemist",
    "chocolate",
    "clothes",
    "coffee",
    "collector",
    "computer",
    "confectionery",
    "convenience",
    "copyshop",
    "cosmetics",
    "craft",
    "curtain",
    "dairy",
    "deli",
    "department_store",
    "doityourself",
    "doors",
    "drugstore",
    "dry_cleaning",
    "e-cigarette",
    "electrical",
    "electronics",
    "energy",
    "erotic",
    "fabric",
    "farm",
    "fashion",
    "fireplace",
    "fishing",
    "flooring",
    "florist",
    "frame",
    "free_flying",
    "frozen_food",
    "fuel",
    "funeral_directors",
    "furniture",
    "games",
    "garden_centre",
    "garden_furniture",
    "gas",
    "general",
    "gift",
    "glaziery",
    "golf",
    "greengrocer",
    "hairdresser",
    "hairdresser_supply",
    "hardware",
    "health_food",
    "hearing_aids",
    "herbalist",
    "hifi",
    "houseware",
    "hunting",
    "ice_cream",
    "interior_decoration",
    "jetski",
    "jewelry",
    "kiosk",
    "kitchen",
    "lamps",
    "laundry",
    "leather",
    "lighting",
    "locksmith",
    "lottery",
    "mall",
    "massage",
    "medical_supply",
    "military_surplus",
    "mobile_phone",
    "model",
    "money_lender",
    "motorcycle",
    "music",
    "musical_instrument",
    "name",
    "newsagent",
    "nutrition_supplements",
    "optician",
    "organic",
    "outdoor",
    "outpost",
    "paint",
    "party",
    "pasta",
    "pastry",
    "pawnbroker",
    "perfumery",
    "pest_control",
    "pet",
    "pet_grooming",
    "phone",
    "photo",
    "pyrotechnics",
    "radiotechnics",
    "religion",
    "robot",
    "scuba_diving",
    "seafood",
    "second_hand",
    "security",
    "sewing",
    "shoes",
    "shop",
    "ski",
    "snowmobile",
    "spices",
    "sports",
    "stationery",
    "storage_rental",
    "supermarket",
    "swimming_pool",
    "tailor",
    "tattoo",
    "tea",
    "ticket",
    "tiles",
    "tobacco",
    "toys",
    "trade",
    "trailer",
    "travel_agency",
    "trophy",
    "tyres",
    "user defined",
    "vacant",
    "vacuum_cleaner",
    "variety_store",
    "video",
    "video_games",
    "watches",
    "water",
    "weapons",
    "wholesale",
    "window_blind",
    "windows",
    "wine",
    "wool",
]

# ========================
# CRAFT TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key%3Acraft
craft_columns = basic_info_tags + [
    "agricultural_engines",
    "atelier",
    "bakery",
    "basket_maker",
    "beekeeper",
    "blacksmith",
    "boatbuilder",
    "bookbinder",
    "brewery",
    "builder",
    "cabinet_maker",
    "carpenter",
    "carpet_layer",
    "caterer",
    "chimney_sweeper",
    "clockmaker",
    "confectionery",
    "cooper",
    "craft",
    "dental_technician",
    "distillery",
    "door_construction",
    "dressmaker",
    "electronics_repair",
    "embroiderer",
    "electrician",
    "engraver",
    "floorer",
    "gardener",
    "glaziery",
    "grinding_mill",
    "handicraft",
    "hvac",
    "insulation",
    "jeweller",
    "joiner",
    "key_cutter",
    "locksmith",
    "metal_construction",
    "mint",
    "musical_instrument",
    "oil_mill",
    "optician",
    "organ_builder",
    "painter",
    "parquet_layer",
    "photographer",
    "photographic_laboratory",
    "piano_tuner",
    "plasterer",
    "plumber",
    "pottery",
    "printer",
    "printmaker",
    "rigger",
    "roofer",
    "saddler",
    "sailmaker",
    "sawmill",
    "scaffolder",
    "sculptor",
    "shoemaker",
    "signmaker",
    "stand_builder",
    "stonemason",
    "sun_protection",
    "tailor",
    "tiler",
    "tinsmith",
    "toolmaker",
    "turner",
    "upholsterer",
    "watchmaker",
    "water_well_drilling",
    "window_construction",
    "winery",
]

# ========================
# LEISURE TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key%3Acraft
leisure_columns = basic_info_tags + [
    "adult_gaming_centre",
    "amusement_arcade",
    "bandstand",
    "beach_resort",
    "bird_hide",
    "bowling_alley",
    "common",
    "dance",
    "disc_golf_course",
    "dog_park",
    "escape_game",
    "firepit",
    "fishing",
    "fitness_centre",
    "fitness_station",
    "garden",
    "golf_course",
    "hackerspace",
    "horse_riding",
    "ice_rink",
    "leisure",
    "marina",
    "miniature_golf",
    "nature_reserve",
    "outdoor_seating",
    "park",
    "picnic_table",
    "pitch",
    "playground",
    "sauna",
    "slipway",
    "sports_centre",
    "stadium",
    "summer_camp",
    "swimming_area",
    "swimming_pool",
    "track",
    "trampoline_park",
    "water_park",
    "wildlife_hide",
]

# ========================
# TOURISM TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key%3Atourism
tourism_columns = basic_info_tags + [
    "alpine_hut",
    "apartment",
    "aquarium",
    "artwork",
    "attraction",
    "camp_pitch",
    "camp_site",
    "caravan_site",
    "chalet",
    "gallery",
    "guest_house",
    "hostel",
    "hotel",
    "information",
    "motel",
    "museum",
    "picnic_site",
    "theme_park",
    "tourism",
    "viewpoint",
    "wilderness_hut",
    "zoo",
]

# ========================
# HISTORIC TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key%3Ahistoric
historic_columns = basic_info_tags + [
    "aircraft",
    "aqueduct",
    "archaeological_site",
    "battlefield",
    "boundary_stone",
    "building",
    "cannon",
    "castle",
    "castle_wall",
    "church",
    "city_gate",
    "citywalls",
    "farm",
    "fort",
    "gallows",
    "highwater_mark",
    "historic",
    "locomotive",
    "manor",
    "memorial",
    "milestone",
    "monastery",
    "monument",
    "optical_telegraph",
    "pillory",
    "railway_car",
    "ruins",
    "rune_stone",
    "ship",
    "tank",
    "tomb",
    "tower",
    "wayside_cross",
    "wayside_shrine",
    "wreck",
]

# ========================
# EMERGENCY TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key%3Aemergency
emergency_columns = basic_info_tags + [
    "access_point",
    "ambulance_station",
    "assembly_point",
    "defibrillator",
    "drinking_water",
    "dry_riser_inlet",
    "emergency",
    "emergency_ward_entrance",
    "fire_alarm_box",
    "fire_extinguisher",
    "fire_flapper",
    "fire_hose",
    "fire_hydrant",
    "fire_water_pond",
    "first_aid_kit",
    "landing_site",
    "life_ring",
    "lifeguard",
    "lifeguard_base",
    "lifeguard_platform",
    "lifeguard_tower",
    "mountain_rescue",
    "phone",
    "rescue_box",
    "siren",
    "suction_point",
    "water_tank",
]

# ========================
# OFFICE TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key%3Aoffice

office_columns = basic_info_tags + [
    "accountant",
    "adoption_agency",
    "advertising_agency",
    "architect",
    "association",
    "bail_bond_agent",
    "charity",
    "company",
    "consulting",
    "coworking",
    "diplomatic",
    "educational_institution",
    "employment_agency",
    "energy_supplier",
    "engineer",
    "estate_agent",
    "financial",
    "forestry",
    "foundation",
    "geodesist",
    "government",
    "graphic_design",
    "guide",
    "harbour_master",
    "insurance",
    "it",
    "lawyer",
    "logistics",
    "moving_company",
    "newspaper",
    "ngo",
    "notary",
    "office",
    "political_party",
    "private_investigator",
    "property_management",
    "quango",
    "religion",
    "research",
    "security",
    "surveyor",
    "tax",
    "tax_advisor",
    "telecommunication",
    "union",
    "visa",
    "water_utility",
]

# ========================
# LANDUSE TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key:landuse
landuse_columns = [
    "allotments",
    "basin",
    "brownfield",
    "cemetery",
    "commercial",
    "conservation",
    "construction",
    "depot",
    "farmland",
    "farmyard",
    "forest",
    "garages",
    "grass",
    "greenfield",
    "greenhouse_horticulture",
    "industrial",
    "landfill",
    "landuse",
    "meadow",
    "military",
    "orchard",
    "peat_cutting",
    "plant_nursery",
    "port",
    "quarry",
    "railway",
    "recreation_ground",
    "religious",
    "reservoir",
    "residential",
    "retail",
    "salt_pond",
    "village_green",
    "vineyard",
    "yes",
]

# ========================
# NATURAL TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key%3Anatural
natural_columns = [
    "arete",
    "bare_rock",
    "bay",
    "beach",
    "cape",
    "cave_entrance",
    "cliff",
    "coastline",
    "dune",
    "fell",
    "geyser",
    "glacier",
    "grassland",
    "heath",
    "hill",
    "hot_spring",
    "isthmus",
    "moor",
    "mud",
    "natural",
    "peak",
    "peninsula",
    "reef",
    "ridge",
    "rock",
    "saddle",
    "sand",
    "scree",
    "scrub",
    "shingle",
    "sinkhole",
    "spring",
    "stone",
    "strait",
    "tree",
    "tree_row",
    "valley",
    "volcano",
    "water",
    "wetland",
    "wood",
]

# ========================
# WATERWAY TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key%3Awaterway
waterway_columns = [
    "boatyard",
    "canal",
    "dam",
    "ditch",
    "dock",
    "drain",
    "drystream",
    "fairway",
    "fish_pass",
    "fuel",
    "lock_gate",
    "pressurised",
    "river",
    "riverbank",
    "stream",
    "tidal_channel",
    "turning_point",
    "wadi",
    "water_point",
    "waterfall",
    "waterway",
    "weir",
]

# ========================
# GEOLOGICAL TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key%3Ageological
geological_columns = [
    "geological",
    "moraine",
    "outcrop",
    "palaeontological_site",
]

# ========================
# POWER TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key%3Apower
power_columns = [
    "cable",
    "catenary_mast",
    "compensator",
    "converter",
    "generator",
    "heliostat",
    "insulator",
    "line",
    "busbar",
    "bay",
    "minor_line",
    "plant",
    "pole",
    "portal",
    "power",
    "substation",
    "switch",
    "switchgear",
    "terminal",
    "tower",
    "transformer",
]

# ========================
# PUBLIC_TRANSPORT TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key%3Apublic_transport
public_transport_columns = basic_info_tags + [
    "stop_position",
    "platform",
    "public_transport",
    "station",
    "stop_area",
]

# ========================
# RAILWAY TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key:railway
railway_columns = [
    "abandoned",
    "buffer_stop",
    "construction",
    "crossing",
    "derail",
    "disused",
    "funicular",
    "halt",
    "level_crossing",
    "light_rail",
    "miniature",
    "monorail",
    "narrow_gauge",
    "platform",
    "preserved",
    "rail",
    "railway",
    "railway_crossing",
    "roundhouse",
    "signal",
    "station",
    "subway",
    "subway_entrance",
    "switch",
    "tram",
    "tram_stop",
    "traverser",
    "turntable",
    "wash",
]

# ========================
# ROUTE TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key:route
route_columns = [
    "bicycle",
    "bus",
    "canoe",
    "detour",
    "duration",
    "evacuation",
    "ferry",
    "foot",
    "from",
    "hiking",
    "horse",
    "inline_skates",
    "light_rail",
    "mtb",
    "network",
    "piste",
    "railway",
    "road",
    "route",
    "running",
    "ski",
    "subway",
    "to",
    "train",
    "tracks",
    "tram",
    "trolleybus",
    "type",
]

# ========================
# PLACE TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key%3Aplace
place_columns = [
    "allotments",
    "archipelago",
    "borough",
    "city",
    "city_block",
    "continent",
    "country",
    "county",
    "district",
    "farm",
    "hamlet",
    "island",
    "islet",
    "isolated_dwelling",
    "locality",
    "municipality",
    "neighbourhood",
    "ocean",
    "place",
    "plot",
    "province",
    "quarter",
    "region",
    "sea",
    "square",
    "state",
    "suburb",
    "town",
    "village",
]

# ========================
# AERIALWAY TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key%3Aaerialway
aerialway_columns = [
    "aerialway",
    "cable_car",
    "gondola",
    "mixed_lift",
    "chair_lift",
    "drag_lift",
    "t-bar",
    "j-bar",
    "platter",
    "rope_tow",
    "magic_carpet",
    "zip_line",
    "goods",
    "pylon",
    "station",
]

# ========================
# AEROWAY TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key%3Aaeroway
aeroway_columns = [
    "aerodrome",
    "aeroway",
    "apron",
    "control_tower",
    "control_center",
    "gate",
    "hangar",
    "helipad",
    "heliport",
    "navigationaid",
    "beacon",
    "runway",
    "taxilane",
    "taxiway",
    "terminal",
    "windsock",
    "highway_strip",
]


# ========================
# BOUNDARY TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key:boundary
boundary_columns = basic_info_tags + [
    "aboriginal_lands",
    "admin_level",
    "administrative",
    "boundary",
    "left:district",
    "right:district",
    "maritime",
    "marker",
    "national_park",
    "political",
    "postal_code",
    "protected_area",
    "admin_level",
    "border_type",
    "start_date",
]



from pyrosm.data_manager import get_osm_data
from pyrosm.frames import prepare_geodataframe
from pyrosm.utils import validate_custom_filter
import geopandas as gpd
import warnings


def get_boundary_data(
    node_coordinates,
    way_records,
    relations,
    tags_as_columns,
    custom_filter,
    boundary_type,
    name,
    bounding_box,
):

    if boundary_type == "all":
        boundary_type = True
    else:
        boundary_type = [boundary_type]

    # If custom_filter has not been defined, initialize with default
    if custom_filter is None:
        custom_filter = {"boundary": boundary_type}

    if "boundary" not in custom_filter.keys():
        custom_filter["boundary"] = True

    # Check that the custom filter is in correct format
    custom_filter = validate_custom_filter(custom_filter)

    # Call signature for fetching buildings
    nodes, ways, relation_ways, relations = get_osm_data(
        node_arrays=None,
        way_records=way_records,
        relations=relations,
        tags_as_columns=tags_as_columns,
        data_filter=custom_filter,
        filter_type="keep",
        osm_keys=None,
    )

    # If there weren't any data, return empty GeoDataFrame
    if nodes is None and ways is None and relations is None:
        warnings.warn(
            "Could not find any boundaries for given area.", UserWarning, stacklevel=2
        )
        return None

    # Prepare GeoDataFrame
    gdf = prepare_geodataframe(
        nodes,
        node_coordinates,
        ways,
        relations,
        relation_ways,
        tags_as_columns,
        bounding_box,
    )

    if gdf is None:
        return None

    # Filter by name
    # (use Pandas for filtering, which allows using 'contains' more easily)
    if name is not None:
        if "name" not in gdf.columns:
            raise ValueError(
                "Could not filter by name from given area. "
                "Any of the OSM elements did not have a name tag."
            )
        gdf = gdf.dropna(subset=["name"])
        gdf = gdf.loc[gdf["name"].str.contains(name)].reset_index(drop=True).copy()

    return gdf



from pyrosm.utils._compat import HAS_IGRAPH, HAS_NETWORKX, HAS_PANDANA
from pyrosm.config import Conf
import geopandas as gpd
import pandas as pd
import numpy as np

# The values used to determine oneway road in OSM
oneway_values = Conf.oneway_values

cpdef _create_igraph(nodes,
                     edges,
                     from_id_col,
                     to_id_col,
                     node_id_col):
    """
    Creates a iGraph from directed edges and nodes.
    NOTE: Assumes that the input edges GeoDataFrame is directed.  
    """
    if not HAS_IGRAPH:
        raise ImportError("'python-igraph' needs to be installed "
                          "in order to export the network for igraph.")
    import igraph

    cdef long long i

    nodes = nodes.copy()
    edges = edges.copy()

    from_id_int = from_id_col + "_seq"
    to_id_int = to_id_col + "_seq"

    edge_list = []

    n_edges = len(edges)
    n_nodes = len(nodes)

    # Add columns for sequential ids
    edges[from_id_int] = None
    edges[to_id_int] = None

    # Convert edges to dict
    edges = edges.to_dict(orient="list")

    # Node-ids needs to be sequential for igraph
    nodes = nodes.reset_index(drop=True)
    nodes["node_id"] = nodes.index

    # Prepare dictionary for fast lookups
    node_dict = {k: v for k, v in zip(nodes[node_id_col].to_list(), nodes["node_id"].to_list())}

    # Node attributes
    node_attributes = nodes.to_dict(orient='list')

    # Generate edge dictionary
    for i in range(0, n_edges):

        # Get nodeids for the edge
        # ------------------------
        # Note: In some cases the node for from/to_id might not exist
        # on the "edge" of the network (e.g. if data has been cropped manually).
        try:
            from_node_id = edges[from_id_col][i]
            from_seq_id = node_dict[from_node_id]
        except KeyError:
            continue
        except Exception as e:
            raise e

        try:
            to_node_id = edges[to_id_col][i]
            to_seq_id = node_dict[to_node_id]
        except KeyError:
            continue
        except Exception as e:
            raise e

        # Add sequential ids to edge_list
        edge_list.append([from_seq_id, to_seq_id])

        # Update the edge attributes
        edges[from_id_int][i] = from_seq_id
        edges[to_id_int][i] = to_seq_id

    del node_dict

    # Create directed graph
    graph = igraph.Graph(n=n_nodes, directed=True, edges=edge_list,
                         vertex_attrs=node_attributes,
                         edge_attrs=edges)
    return graph

cpdef _create_nxgraph(nodes,
                      edges,
                      from_id_col,
                      to_id_col,
                      node_id_col):
    """
    Creates a NetworkX graph from directed edges and nodes.
    NOTE: Assumes that the input edges GeoDataFrame is directed.
    """
    if not HAS_NETWORKX:
        raise ImportError("'networkx' needs to be installed "
                          "in order to export the network for networkx / osmnx.")
    import networkx as nx

    cdef long long i

    nodes = nodes.copy()
    edges = edges.copy()

    crs = f"EPSG:{edges.crs.to_epsg(min_confidence=25)}"
    n_edges = len(edges)
    edge_list = []

    # Convert edges to dict
    edge_attributes = edges.to_dict(orient="index")

    # Prepare node dictionary for fast lookups
    node_dict = {k: None for k in nodes[node_id_col].to_list()}

    # Node attributes
    node_attributes = nodes.to_dict(orient="index")
    node_attributes = [(k, v) for k, v in node_attributes.items()]

    # Generate edge dictionary
    for i in range(0, n_edges):

        # Get nodeids for the edge
        # ------------------------
        # Note: In some cases the node for from/to_id might not exist
        # on the "edge" of the network (e.g. if data has been cropped manually).
        try:
            from_node_id = edge_attributes[i][from_id_col]

            # Check if the data for node exists
            node_dict[from_node_id]

        except KeyError:
            continue
        except Exception as e:
            raise e

        try:
            to_node_id = edge_attributes[i][to_id_col]

            # Check if the data for node exists
            node_dict[to_node_id]
        except KeyError:
            continue
        except Exception as e:
            raise e

        # Create edges
        # ------------
        edge_list.append([from_node_id, to_node_id, 0, edge_attributes[i]])

    del node_dict

    # Create directed graph
    graph = nx.MultiDiGraph()
    graph.add_nodes_from(node_attributes)
    graph.add_edges_from(edge_list)
    graph.graph["crs"] = crs
    graph.graph["name"] = "Made with Pyrosm library."

    return graph

cpdef _create_pdgraph(nodes,
                      edges,
                      from_id_col,
                      to_id_col,
                      weight_cols):
    """
    Creates a Pandana Network from directed edges and nodes.
    NOTE: Assumes that the input edges GeoDataFrame is directed.
    """
    if not HAS_PANDANA:
        raise ImportError("'pandana' needs to be installed "
                          "in order to export the network for it.")
    from pandana import Network
    return Network(node_x=nodes["x"],
                   node_y=nodes["y"],
                   edge_from=edges[from_id_col],
                   edge_to=edges[to_id_col],
                   edge_weights=edges[weight_cols],
                   twoway=False)


cpdef generate_directed_edges(edges,
                              direction,
                              from_id_col,
                              to_id_col,
                              force_bidirectional):
    """
    Generates directed set of edges from network 
    following rules specified in 'direction' column.
    
    If 'force_bidirectional=True' travel to both direction is allowed for all edges.
    """
    if force_bidirectional:
        # Flip from/to values
        edges_dir2 = edges.copy(deep=True).rename(
            columns={to_id_col: from_id_col,
                     from_id_col: to_id_col}
        )
        return pd.concat([edges, edges_dir2], ignore_index=True)

    # ========================================
    # Directed edges according 'oneway' rules
    # ========================================

    if "junction" in edges.columns:
        roundabouts = True
    else:
        roundabouts = False

    if roundabouts:
        # Edge is oneway if it is tagged as such OR if it tagged as roundabout
        oneway_mask = (edges[direction].isin(oneway_values)) | (edges["junction"] == "roundabout")
    else:
        oneway_mask = edges[direction].isin(oneway_values)

    edge_cnt = len(edges)
    oneway_edges = edges.loc[oneway_mask].copy()
    twoway_edges = edges.loc[~oneway_mask].copy()
    twoway_edges_dir2 = twoway_edges.copy(deep=True).rename(columns={to_id_col: from_id_col, from_id_col: to_id_col})
    twoway_edges_dir2.index = np.arange(edge_cnt, edge_cnt + len(twoway_edges))

    # Select edges that are allowed only to opposite direction
    against_mask = oneway_edges[direction].isin(["-1", "T"])
    against_edges = oneway_edges.loc[against_mask].copy()
    along_edges = oneway_edges.loc[~against_mask].copy()  # Nothing needs to be done for these

    # Flip the from/to ids for against edges
    against_edges = against_edges.rename(columns={from_id_col: to_id_col, to_id_col: from_id_col})

    # Stack everything (keep order)
    return pd.concat([along_edges, against_edges, twoway_edges, twoway_edges_dir2])



from collections import defaultdict


def _get_node_successors(edges, from_id_col, to_id_col):
    edge_cnt = len(edges)
    node_successors = defaultdict(list)
    from_ids = edges[from_id_col].to_list()
    to_ids = edges[to_id_col].to_list()

    for i in range(0, edge_cnt):
        node_successors[from_ids[i]].append(to_ids[i])
    return node_successors


def _strongly_connected_components(list_of_nodes, node_successors):
    """
    Generate nodes in strongly connected components of graph.

    Source: https://networkx.org/documentation/stable/reference/algorithms/component.html

    Uses Tarjan's algorithm [1] with Nuutila's modifications [2].
    Nonrecursive version of algorithm.

    References
    ----------

    [1] Depth-first search and linear graph algorithms, R. Tarjan
        SIAM Journal of Computing 1(2):146-160, (1972).

    [2] On finding the strongly connected components in a directed graph.
        E. Nuutila and E. Soisalon-Soinen
        Information Processing Letters 49(1): 9-14, (1994).

    """
    preorder = {}
    lowlink = {}
    scc_found = {}
    scc_queue = []
    i = 0  # Preorder counter
    for source in list_of_nodes:
        if source not in scc_found:
            queue = [source]
            while queue:
                v = queue[-1]
                if v not in preorder:
                    i = i + 1
                    preorder[v] = i
                done = 1
                v_nbrs = node_successors[v]
                for w in v_nbrs:
                    if w not in preorder:
                        queue.append(w)
                        done = 0
                        break
                if done == 1:
                    lowlink[v] = preorder[v]
                    for w in v_nbrs:
                        if w not in scc_found:
                            if preorder[w] > preorder[v]:
                                lowlink[v] = min([lowlink[v], lowlink[w]])
                            else:
                                lowlink[v] = min([lowlink[v], preorder[w]])
                    queue.pop()
                    if lowlink[v] == preorder[v]:
                        scc_found[v] = True
                        scc = {v}
                        while scc_queue and preorder[scc_queue[-1]] > preorder[v]:
                            k = scc_queue.pop()
                            scc_found[k] = True
                            scc.add(k)
                        yield scc
                    else:
                        scc_queue.append(v)


def get_connected_edges(nodes, edges, from_id_col="u", to_id_col="v", node_id_col="id"):
    """Filters the network data (directed) to include only connected edges and nodes."""
    node_successors = _get_node_successors(edges, from_id_col, to_id_col)
    node_ids = nodes[node_id_col].to_list()
    scc = max(_strongly_connected_components(node_ids, node_successors), key=len)
    # Filter nodes and edges accordingly
    n = nodes[nodes[node_id_col].isin(scc)]
    e = edges[(edges[from_id_col].isin(scc)) & (edges[to_id_col].isin(scc))]
    return n, e.reset_index(drop=True)



from pyrosm.data_manager import get_osm_data
from pyrosm.frames import prepare_geodataframe
import warnings


def get_network_data(
    node_coordinates,
    way_records,
    tags_as_columns,
    network_filter,
    bounding_box,
    slice_to_segments,
):
    # Tags to keep as separate columns
    tags_as_columns += ["id", "nodes", "timestamp", "changeset", "version"]

    # Call signature for fetching network data
    nodes, ways, relation_ways, relations = get_osm_data(
        node_arrays=None,
        way_records=way_records,
        relations=None,
        tags_as_columns=tags_as_columns,
        data_filter=network_filter,
        filter_type="exclude",
        # Keep only records having 'highway' tag
        osm_keys="highway",
    )

    # If there weren't any data, return None
    if ways is None:
        warnings.warn(
            "Could not find any edges for given area.", UserWarning, stacklevel=2
        )
        return None, None

    # Prepare GeoDataFrame
    edges, nodes = prepare_geodataframe(
        nodes,
        node_coordinates,
        ways,
        relations,
        relation_ways,
        tags_as_columns,
        bounding_box,
        parse_network=True,
        calculate_seg_lengths=slice_to_segments,
    )

    return edges, nodes



from enum import Enum
import numpy as np
from pygeos.coordinates import get_coordinates


class Unit(Enum):
    """
    Enumeration of supported units.
    The full list can be checked by iterating over the class; e.g.
    the expression `tuple(Unit)`.

    Inspired by: https://github.com/mapado/haversine

    """

    KILOMETERS = "km"
    METERS = "m"
    MILES = "mi"
    NAUTICAL_MILES = "nmi"
    FEET = "ft"
    INCHES = "in"


# mean earth radius - https://en.wikipedia.org/wiki/Earth_radius#Mean_radius
_AVG_EARTH_RADIUS_KM = 6371.0088

# Unit values taken from http://www.unitconversion.org/unit_converter/length.html
_CONVERSIONS = {
    Unit.KILOMETERS: 1.0,
    Unit.METERS: 1000.0,
    Unit.MILES: 0.621371192,
    Unit.NAUTICAL_MILES: 0.539956803,
    Unit.FEET: 3280.839895013,
    Unit.INCHES: 39370.078740158,
}


def haversine(lat1, lng1, lat2, lng2, unit=Unit.KILOMETERS):
    """
    Calculate the great-circle distance between two points on the Earth surface.
    Takes two 2-tuples, containing the latitude and longitude of each point in decimal degrees,
    and, optionally, a unit of length.

    Inspired by: https://github.com/mapado/haversine

    Parameters
    ==========
    lat1 : np.array
        Latitude in decimal degrees
    lng1 : np.array
        Longitude in decimal degrees
    lat2 : np.array
        Latitude in decimal degrees
    lng2 : np.arrays
        Longitude in decimal degrees
    unit : pyrosm.distance.Unit
        A member of pyrosm.distance.Unit, or, equivalently, a string containing the
        initials of its corresponding unit of measurement (i.e. miles = mi)
        default 'km' (kilometers).
    """

    # get earth radius in required units
    unit = Unit(unit)
    avg_earth_radius = _AVG_EARTH_RADIUS_KM * _CONVERSIONS[unit]

    # convert all latitudes/longitudes from decimal degrees to radians
    lat1, lng1, lat2, lng2 = map(np.deg2rad, (lat1, lng1, lat2, lng2))

    # calculate haversine
    lat = lat2 - lat1
    lng = lng2 - lng1
    d = np.sin(lat * 0.5) ** 2 + np.cos(lat1) * np.cos(lat2) * np.sin(lng * 0.5) ** 2

    return 2 * avg_earth_radius * np.arcsin(np.sqrt(d))


def calculate_geom_length(geom):
    return calculate_geom_array_length(geom).sum().round(0)


def calculate_geom_array_length(geom_array):
    coords = get_coordinates(geom_array).T

    # Only every second element should be taken from the coordinates
    lon1, lat1 = coords[0][:-1:2], coords[1][:-1:2]
    lon2, lat2 = coords[0][1::2], coords[1][1::2]

    # Length of the segments
    geom_lengths = haversine(lat1, lon1, lat2, lon2, unit=Unit.METERS).round(3)
    return geom_lengths



from pyrosm.graph_export import (
    _create_igraph,
    _create_nxgraph,
    _create_pdgraph,
    generate_directed_edges,
)
from pyrosm.graph_connectivity import get_connected_edges
from pyrosm.utils import validate_edge_gdf, validate_node_gdf
from pyrosm.config import Conf
import warnings


def get_directed_edges(
    nodes,
    edges,
    direction="oneway",
    from_id_col="u",
    to_id_col="v",
    node_id_col="id",
    force_bidirectional=False,
    network_type=None,
):
    """Prepares the edges and nodes for exporting to different graphs."""
    allowed_network_types = Conf._possible_network_filters

    # Validate nodes and edges
    validate_node_gdf(nodes)
    validate_edge_gdf(edges)

    for col in [from_id_col, to_id_col]:
        if col not in edges.columns:
            raise ValueError(
                "Required column '{col}' does not exist in edges.".format(col=col)
            )

    if direction not in edges.columns:
        warnings.warn(
            f"Column '{direction}' missing in the edges GeoDataFrame. "
            f"Assuming all edges to be bidirectional "
            f"(travel allowed to both directions).",
            UserWarning,
            stacklevel=2,
        )
        edges[direction] = None

    if node_id_col not in nodes.columns:
        raise ValueError(
            "Required column '{col}' does not exist in nodes.".format(col=node_id_col)
        )

    # Check the network_type
    if network_type is not None:
        net_type = network_type
    else:
        net_type = edges._metadata[-1]

    # Check if network type is stored with edges or nodes
    if net_type not in allowed_network_types:
        net_type = nodes._metadata[-1]
        if net_type not in allowed_network_types:
            txt = ", ".join(allowed_network_types)
            raise ValueError(
                "Could not detect the network type from the edges. "
                "In order to save the graph, specify the type of your network"
                "with 'network_type' -parameter."
                "Possible network types are: " + txt
            )

    edges = edges.copy()
    nodes = nodes.copy()

    # Generate directed edges
    # Check if user wants to force bidirectional graph
    # or if the graph is walking, cycling or all
    if force_bidirectional or net_type in ["walking", "cycling", "all"]:
        edges = generate_directed_edges(
            edges, direction, from_id_col, to_id_col, force_bidirectional=True
        )
    else:
        edges = generate_directed_edges(
            edges, direction, from_id_col, to_id_col, force_bidirectional=False
        )

    return nodes, edges


def to_networkx(
    nodes,
    edges,
    direction="oneway",
    from_id_col="u",
    to_id_col="v",
    edge_id_col="id",
    node_id_col="id",
    force_bidirectional=False,
    network_type=None,
    retain_all=False,
    osmnx_compatible=True,
):
    """
    Creates a NetworkX.MultiDiGraph from given OSM GeoDataFrame.

    Parameters
    ----------
    edges : GeoDataFrame
        GeoDataFrame containing road network data.

    network_type : str
        The type of the network. Possible values:
              - `'walking'`
              - `'cycling'`
              - `'driving'`
              - `'driving+service'`
              - `'all'`.

    direction : str
        Name for the column containing information about the allowed driving directions

    from_id_col : str
        Name for the column having the from-node-ids of edges.

    to_id_col : str
        Name for the column having the to-node-ids of edges.

    edge_id_col : str
        Name for the column having the unique id for edges.

    node_id_col : str
        Name for the column having the unique id for nodes.

    force_bidirectional : bool
        If True, all edges will be created as bidirectional (allow travel to both directions).

    network_type : str
        Network type for the given data. Determines how the graph will be constructed.
        By default, bidirectional graph is created for walking, cycling and all,
        and directed graph for driving (i.e. oneway streets are taken into account).
        Possible values are: 'walking', 'cycling', 'driving', 'driving+service', 'all'.

    retain_all : bool
        if True, return the entire graph even if it is not connected.
        otherwise, retain only the connected edges.

    osmnx_compatible : bool (default True)
        if True, modifies the edge and node-attribute naming to be compatible with OSMnx
        (allows utilizing all OSMnx functionalities).

    Returns
    -------
    networkx.MultiDiGraph

    """

    # Prepare the data
    nodes, edges = get_directed_edges(
        nodes,
        edges,
        direction,
        from_id_col,
        to_id_col,
        node_id_col,
        force_bidirectional,
        network_type,
    )

    # Keep only strongly connected component if not specifically requested otherwise
    if not retain_all:
        nodes, edges = get_connected_edges(
            nodes, edges, from_id_col, to_id_col, node_id_col
        )

    if osmnx_compatible:
        # add 'key' attribute which is needed by OSMnx
        if "key" not in edges.columns:
            edges["key"] = 0

        # Follow the naming convention of OSMnx
        nodes = nodes.rename(columns={node_id_col: "osmid", "lat": "y", "lon": "x"})
        edges = edges.rename(columns={edge_id_col: "osmid"})
        node_id_col = "osmid"

    # Add node-id as index
    nodes = nodes.set_index(node_id_col, drop=False)
    nodes = nodes.rename_axis([None])

    # Create NetworkX graph
    return _create_nxgraph(nodes, edges, from_id_col, to_id_col, node_id_col)


def to_igraph(
    nodes,
    edges,
    direction="oneway",
    from_id_col="u",
    to_id_col="v",
    node_id_col="id",
    force_bidirectional=False,
    network_type=None,
    retain_all=False,
):
    """
    Creates an iGraph from given OSM GeoDataFrame.

    Parameters
    ----------
    edges : GeoDataFrame
        GeoDataFrame containing road network data.

    network_type : str
        The type of the network. Possible values:
              - `'walking'`
              - `'cycling'`
              - `'driving'`
              - `'driving+service'`
              - `'all'`.

    direction : str
        Name for the column containing information about the allowed driving directions

    from_id_col : str
        Name for the column having the from-node-ids of edges.

    to_id_col : str
        Name for the column having the to-node-ids of edges.

    edge_id_col : str
        Name for the column having the unique id for edges.

    node_id_col : str
        Name for the column having the unique id for nodes.

    force_bidirectional : bool
        If True, all edges will be created as bidirectional (allow travel to both directions).

    network_type : str
        Network type for the given data. Determines how the graph will be constructed.
        By default, bidirectional graph is created for walking, cycling and all,
        and directed graph for driving (i.e. oneway streets are taken into account).
        Possible values are: 'walking', 'cycling', 'driving', 'driving+service', 'all'.

    retain_all : bool
        if True, return the entire graph even if it is not connected.
        otherwise, retain only the connected edges.

    Returns
    -------
    igraph.Graph

    """
    # Prepare the data
    nodes, edges = get_directed_edges(
        nodes,
        edges,
        direction,
        from_id_col,
        to_id_col,
        node_id_col,
        force_bidirectional,
        network_type,
    )

    # Keep only strongly connected component if not specifically requested otherwise
    if not retain_all:
        nodes, edges = get_connected_edges(
            nodes, edges, from_id_col, to_id_col, node_id_col
        )

    return _create_igraph(nodes, edges, from_id_col, to_id_col, node_id_col)


def to_pandana(
    nodes,
    edges,
    direction="oneway",
    from_id_col="u",
    to_id_col="v",
    node_id_col="id",
    force_bidirectional=False,
    network_type=None,
    retain_all=False,
    weight_cols=["length"],
):
    # Prepare the data
    nodes, edges = get_directed_edges(
        nodes,
        edges,
        direction,
        from_id_col,
        to_id_col,
        node_id_col,
        force_bidirectional,
        network_type,
    )

    # Keep only strongly connected component if not specifically requested otherwise
    if not retain_all:
        nodes, edges = get_connected_edges(
            nodes, edges, from_id_col, to_id_col, node_id_col
        )

    nodes = nodes.rename(columns={"lat": "y", "lon": "x"})
    nodes = nodes.set_index("id", drop=False)
    nodes = nodes.rename_axis([None])

    return _create_pdgraph(nodes, edges, from_id_col, to_id_col, weight_cols)
